// dedicated to Andrew Ladd
// code by jph

#define           jDRP_type     float

// public
int               jDRP_w;               // read only
int               jDRP_h;               // read only
linked<int>       jDRP_plan;            // read only

// private
jDRP_type*        jDRP_dens;
jDRP_type*        jDRP_dist;
int*              jDRP_pred;
int*              jDRP_mask;
SDL_Surface*      jDRP_image;
SDL_Surface*      jDRP_route;
int               jDRP_numel;
SDL_PixelFormat   jDRP_PixelFormat;           // pixel format for all SDL_Surfaces that we create


void jDRP_protected(void) {
  jDRP_PixelFormat.palette=NULL;
  jDRP_PixelFormat.BitsPerPixel=32;
  jDRP_PixelFormat.BytesPerPixel=4;
  jDRP_PixelFormat.Amask=0xff000000;
  jDRP_PixelFormat.Rmask=0x00ff0000;
  jDRP_PixelFormat.Gmask=0x0000ff00;
  jDRP_PixelFormat.Bmask=0x000000ff;  
  jDRP_PixelFormat.Ashift=24;
  jDRP_PixelFormat.Rshift=16;
  jDRP_PixelFormat.Gshift=8;
  jDRP_PixelFormat.Bshift=0;
  jDRP_PixelFormat.Aloss=0;
  jDRP_PixelFormat.Rloss=0;
  jDRP_PixelFormat.Gloss=0;
  jDRP_PixelFormat.Bloss=0;  
  jDRP_PixelFormat.colorkey=0;
  jDRP_PixelFormat.alpha=0;
}

void jDRP_constructor(char* fid) {
  int c1;
  
  jDRP_protected();
  
  SDL_Surface* mySDL_Surface;
  mySDL_Surface=SDL_LoadBMP(fid);
    jDRP_image=SDL_ConvertSurface(mySDL_Surface,&jDRP_PixelFormat,SDL_SWSURFACE);
    jDRP_route=SDL_ConvertSurface(mySDL_Surface,&jDRP_PixelFormat,SDL_SWSURFACE);
  SDL_FreeSurface(mySDL_Surface);

  jDRP_w=jDRP_image->w;
  jDRP_h=jDRP_image->h;
  jDRP_numel=jDRP_w*jDRP_h;
  jDRP_dens=new jDRP_type[jDRP_numel];  
  jDRP_dist=new jDRP_type[jDRP_numel];
  jDRP_pred=new int[jDRP_numel];
  jDRP_mask=new int[jDRP_numel];
  
  SDL_LockSurface(jDRP_image);
  Uint32* myUint32=(Uint32*)(jDRP_image->pixels);
  int myR,myG,myB;  
  for (c1=0;c1<jDRP_numel;c1++) {
    myR=myUint32[c1]&0xff0000;  myR>>=16;
    myG=myUint32[c1]&0xff00;    myG>>=8;
    myB=myUint32[c1]&0xff;
    jDRP_dens[c1]=(jDRP_type)(myR+myG+myB)/(jDRP_type)3;
  }
  SDL_UnlockSurface(jDRP_image);  
}

void jDRP_destructor(void) {
  SDL_FreeSurface(jDRP_image);
  SDL_FreeSurface(jDRP_route);    
  delete[] jDRP_dens;
  delete[] jDRP_dist;
  delete[] jDRP_pred;
  delete[] jDRP_mask;  
}

void jDRP_routeFrom(int pix,int piy,jDRP_type lim) {
  int c1,c2,c3,c4;
  int pin,pia,pii;  
  for (c1=0;c1<jDRP_numel;c1++) {
    jDRP_dist[c1]=jDRP_numel;
    jDRP_pred[c1]=0;
    jDRP_mask[c1]=0;
  }
  // BEGIN: do fancy shortest-path algorithm
  pia=piy*jDRP_w+pix;                               // map the robot into the map  
  linked<int>  list0,list1;
  linked<int>* now=&list0;
  linked<int>* nxt=&list1;
  linked<int>* tmp;
  now->append(pia);
  jDRP_dist[pia]=0;
  jDRP_pred[pia]=pia;
  jDRP_type one=1,rt2=sqrt(2.0),rt5=sqrt(5.0);
  
  //int       ofs[8]={-jDRP_w-1,-jDRP_w+1, jDRP_w+1, jDRP_w-1,  1,-jDRP_w, -1,jDRP_w};
  //jDRP_type fac[8]={ rt2     , rt2     , rt2     , rt2     ,one, one   ,one,one   };

  int       ofs[16]={-jDRP_w-jDRP_w-1,-jDRP_w-jDRP_w+1,-jDRP_w-2,-jDRP_w+2,jDRP_w-2,jDRP_w+2,jDRP_w+jDRP_w-1,jDRP_w+jDRP_w+1, -jDRP_w-1,-jDRP_w+1, jDRP_w+1, jDRP_w-1,  1,-jDRP_w, -1,jDRP_w};
  jDRP_type fac[16]={rt5,rt5,rt5,rt5,rt5,rt5,rt5,rt5, rt2     , rt2     , rt2     , rt2     ,one, one   ,one,one   };

  jDRP_type met;
  int cnt=0;
  while (now->next[0]&&cnt<lim) {                   // repeat the process until now is empty, or mhh...
    cnt++;
    while (now->token(pia)) {                       // in the while loop, pia contains pixel address to be processed
      for (c1=0;c1<16;c1++) {
        pii=pia+ofs[c1];
        if (0<pii&&pii<jDRP_numel)
          if (met=jDRP_dens[pii]*fac[c1])            // only visit explored terrain (metric =0 whereever unexplored)
            if (jDRP_dist[pia]+met<jDRP_dist[pii]&&jDRP_dist[pia]+met<lim) {    // IF ROUTING DOES NOT WORK, REMOVE latter condition!!!
              jDRP_dist[pii]=jDRP_dist[pia]+met;
              jDRP_pred[pii]=pia;
              if (jDRP_mask[pii]<cnt) {
                jDRP_mask[pii]=cnt;
                nxt->append(pii);
              }
            }
      }
    }
    while (now->token(pia))                         // remove
      now->remove();
    tmp=now;
    now=nxt;
    nxt=tmp;
  }
  // END: do fancy shortest-path algorithm
  // BEGIN: for visualization only
  SDL_FillRect(jDRP_route,NULL,SDL_MapRGBA(&jDRP_PixelFormat,000,000,000,000));
  SDL_LockSurface(jDRP_route);
  Uint32* myUint32=(Uint32*)(jDRP_route->pixels);
  for (c1=0;c1<jDRP_numel;c1++)
    if (jDRP_dist[c1]<jDRP_numel) {
      myUint32[c1]  =(int)jDRP_dist[c1];
      myUint32[c1]>>=3;
      myUint32[c1] |=0x60000000;      
    }
  SDL_UnlockSurface(jDRP_route);
  // END: for visualization only
}

jDRP_type jDRP_planTo(int pix,int piy) {
  int c1,tok,pin;
  int pta=piy*jDRP_w+pix;                              // pixel address of target  
  while (jDRP_plan.token(tok))                         // delete old path
    jDRP_plan.remove();
  while (jDRP_pred[pta]&&pta!=jDRP_pred[pta]) {   // hop from predecessor to next predecessor
    jDRP_plan.prepend(pta);
    pta=jDRP_pred[pta];
  }
  return jDRP_dist[pta];
}

void jDRP_draw(SDL_Surface* mySDL_Surface,SDL_Rect &mySDL_Rect) {
  SDL_BlitSurface(jDRP_image,NULL,mySDL_Surface,&mySDL_Rect);
  SDL_BlitSurface(jDRP_route,NULL,mySDL_Surface,&mySDL_Rect);
}
